﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.QuadraticAssignment {
  [Item("AffinitySwapMoveEvaluator", "Evaluated an affinity-based swap move on a QAP solution.")]
  [StorableClass]
  public class AffinitySwapMoveEvaluator : QAPMoveEvaluator, IAffinitySwapMoveOperator {
    public ILookupParameter<AffinitySwapMove> AffinitySwapMoveParameter {
      get { return (ILookupParameter<AffinitySwapMove>)Parameters["AffinitySwapMove"]; }
    }

    [StorableConstructor]
    protected AffinitySwapMoveEvaluator(bool deserializing) : base(deserializing) { }
    protected AffinitySwapMoveEvaluator(AffinitySwapMoveEvaluator original, Cloner cloner)
      : base(original, cloner) {
    }
    public AffinitySwapMoveEvaluator() {
      Parameters.Add(new LookupParameter<AffinitySwapMove>("AffinitySwapMove", "The move to evaluate."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AffinitySwapMoveEvaluator(this, cloner);
    }

    /// <summary>
    /// Calculates the quality of the move <paramref name="move"/> by evaluating the changes.
    /// The runtime complexity of this method is O(M*N) with N being the size of the permutation and M the number of swaps
    /// </summary>
    /// <param name="assignment">The current permutation.</param>
    /// <param name="move">The move that is to be evaluated if it was applied to the current permutation.</param>
    /// <param name="weights">The weights matrix.</param>
    /// <param name="distances">The distances matrix.</param>
    /// <returns>The relative change in quality if <paramref name="move"/> was applied to <paramref name="assignment"/>.</returns>
    public static double Apply(Permutation assignment, AffinitySwapMove move, DoubleMatrix weights, DoubleMatrix distances) {
      if (!move.Swaps.Any()) return 0;
      double moveQuality = 0;

      foreach (KeyValuePair<int, int> swap in move.Swaps) {
        int fac1 = swap.Key, fac2 = swap.Value;
        int loc1 = assignment[fac1], loc2 = assignment[fac2];
        for (int j = 0; j < assignment.Length; j++) {
          if (j == fac1) {
            moveQuality += weights[fac1, fac1] * (distances[loc2, loc2] - distances[loc1, loc1]);
            moveQuality += weights[fac1, fac2] * (distances[loc2, loc1] - distances[loc1, loc2]);
          } else if (j == fac2) {
            moveQuality += weights[fac2, fac2] * (distances[loc1, loc1] - distances[loc2, loc2]);
            moveQuality += weights[fac2, fac1] * (distances[loc1, loc2] - distances[loc2, loc1]);
          } else {
            int locJ = assignment[j];
            bool isMoved = move.Contains(j);
            if (isMoved) {
              moveQuality -= weights[fac1, j] * distances[loc1, locJ];
              moveQuality -= weights[fac2, j] * distances[loc2, locJ];
              locJ = assignment[move.GetSwapLocation(j)];
              moveQuality += weights[fac1, j] * distances[loc2, locJ];
              moveQuality += weights[fac2, j] * distances[loc1, locJ];
            } else {
              moveQuality += weights[fac1, j] * (distances[loc2, locJ] - distances[loc1, locJ]);
              moveQuality += weights[fac2, j] * (distances[loc1, locJ] - distances[loc2, locJ]);
              moveQuality += weights[j, fac1] * (distances[locJ, loc2] - distances[locJ, loc1]);
              moveQuality += weights[j, fac2] * (distances[locJ, loc1] - distances[locJ, loc2]);
            }
          }
        }
      }
      return moveQuality;
    }

    public override IOperation Apply() {
      AffinitySwapMove move = AffinitySwapMoveParameter.ActualValue;
      if (move == null) throw new InvalidOperationException("Swap move is not found.");
      Permutation assignment = PermutationParameter.ActualValue;
      DoubleMatrix distances = DistancesParameter.ActualValue;
      DoubleMatrix weights = WeightsParameter.ActualValue;

      double moveQuality = QualityParameter.ActualValue.Value;
      moveQuality += Apply(assignment, move, weights, distances);
      MoveQualityParameter.ActualValue = new DoubleValue(moveQuality);
      return base.Apply();
    }
  }
}
