﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.QuadraticAssignment {
  [Item("AffinitySwapMove", "An affinity swap move may move affine facilities together with another item.")]
  [StorableClass]
  public class AffinitySwapMove : Item {

    [Storable]
    private Dictionary<int, int> reverseSwaps;
    [Storable]
    private Dictionary<int, int> swaps;
    public IEnumerable<KeyValuePair<int, int>> Swaps {
      get { return swaps.AsEnumerable(); }
    }

    [StorableConstructor]
    protected AffinitySwapMove(bool deserializing) : base(deserializing) { }
    protected AffinitySwapMove(AffinitySwapMove original, Cloner cloner)
      : base(original, cloner) {
      swaps = new Dictionary<int, int>(original.swaps);
      reverseSwaps = new Dictionary<int, int>(original.reverseSwaps);
    }
    public AffinitySwapMove()
      : base() {
      swaps = new Dictionary<int, int>();
      reverseSwaps = new Dictionary<int, int>();
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new AffinitySwapMove(this, cloner);
    }

    public bool AddSwap(int index1, int index2) {
      if (Contains(index1) || Contains(index2)) return false;
      swaps.Add(index1, index2);
      reverseSwaps.Add(index2, index1);
      return true;
    }

    public bool Contains(int index) {
      return swaps.ContainsKey(index) || reverseSwaps.ContainsKey(index);
    }

    /// <summary>
    /// Returns the swap partner for the given index.
    /// </summary>
    /// <remarks>
    /// Throws a KeyNotFoundException when the <paramref name="index"/> is not part of the swap.
    /// Check with <see cref="Contains"/> to see if it is included first.
    /// </remarks>
    /// <param name="index">The index for which to find the swap partner.</param>
    /// <returns>The swap partner of the index.</returns>
    public int GetSwapLocation(int index) {
      if (swaps.ContainsKey(index)) return swaps[index];
      else if (reverseSwaps.ContainsKey(index)) return reverseSwaps[index];
      else throw new KeyNotFoundException(ItemName + ": The given index is not part of the affinity-based swap move.");
    }
  }
}
