﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2011 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.QuadraticAssignment {
  [Item("StochasticAffinitySwapMoveGenerator", "This generator creates moves that also affect affine facilities.")]
  [StorableClass]
  public class StochasticAffinitySwapMoveGenerator : SingleSuccessorOperator, IMoveGenerator, IStochasticOperator, IAffinitySwapMoveOperator {
    public override bool CanChangeName {
      get { return false; }
    }
    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }
    public ILookupParameter<DoubleMatrix> WeightsParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Weights"]; }
    }
    public ILookupParameter<DoubleMatrix> DistancesParameter {
      get { return (ILookupParameter<DoubleMatrix>)Parameters["Distances"]; }
    }
    public ILookupParameter<Permutation> PermutationParameter {
      get { return (ILookupParameter<Permutation>)Parameters["Permutation"]; }
    }
    public ILookupParameter<AffinitySwapMove> AffinitySwapMoveParameter {
      get { return (ILookupParameter<AffinitySwapMove>)Parameters["AffinitySwapMove"]; }
    }
    public ILookupParameter<DoubleArray> WeightsRowSumParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["WeightsRowSum"]; }
    }
    public ILookupParameter<DoubleArray> DistancesRowSumParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["DistancesRowSum"]; }
    }

    [StorableConstructor]
    protected StochasticAffinitySwapMoveGenerator(bool deserializing) : base(deserializing) { }
    protected StochasticAffinitySwapMoveGenerator(StochasticAffinitySwapMoveGenerator original, Cloner cloner) : base(original, cloner) { }
    public StochasticAffinitySwapMoveGenerator() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Weights", "The weights matrix."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Distances", "The distances matrix."));
      Parameters.Add(new LookupParameter<Permutation>("Permutation", "The current assignment."));
      Parameters.Add(new LookupParameter<AffinitySwapMove>("AffinitySwapMove", "The affinity swap moves that are created."));
      Parameters.Add(new LookupParameter<DoubleArray>("WeightsRowSum", "The row sum for each facility denoting the total weight to other facilities."));
      Parameters.Add(new LookupParameter<DoubleArray>("DistancesRowSum", "The row sum for each location denoting the total distance to other locations."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new StochasticAffinitySwapMoveGenerator(this, cloner);
    }

    public override IOperation Apply() {
      IRandom random = RandomParameter.ActualValue;
      DoubleMatrix weights = WeightsParameter.ActualValue;
      DoubleMatrix distances = DistancesParameter.ActualValue;
      DoubleArray weightsRowSum = WeightsRowSumParameter.ActualValue;
      if (weightsRowSum == null) {
        weightsRowSum = GenerateRowSum(weights);
        WeightsRowSumParameter.ActualValue = weightsRowSum;
      }
      DoubleArray distancesRowSum = DistancesRowSumParameter.ActualValue;
      if (distancesRowSum == null) {
        distancesRowSum = GenerateRowSum(distances);
        DistancesRowSumParameter.ActualValue = distancesRowSum;
      }
      Permutation p = PermutationParameter.ActualValue;
      var moves = Generate(random, p, weights, weightsRowSum, distances, distancesRowSum);
      List<Scope> moveScopes = new List<Scope>();
      int i = 1;
      foreach (var move in moves) {
        Scope scope = new Scope(i.ToString());
        scope.Variables.Add(new Variable(AffinitySwapMoveParameter.ActualName, move));
        moveScopes.Add(scope);
      }
      ExecutionContext.Scope.SubScopes.AddRange(moveScopes);
      return base.Apply();
    }

    public static DoubleArray GenerateRowSum(DoubleMatrix matrix) {
      DoubleArray result = new DoubleArray(matrix.Rows);
      for (int i = 0; i < matrix.Rows; i++)
        for (int j = 0; j < matrix.Columns; j++)
          result[i] += matrix[i, j];
      return result;
    }

    public static IEnumerable<AffinitySwapMove> Generate(IRandom random, Permutation permutation, DoubleMatrix weights, DoubleArray weightsRowSum, DoubleMatrix distances, DoubleArray distancesRowSum) {
      int length = permutation.Length;
      if (length == 1) throw new ArgumentException("ExhaustiveSwap2MoveGenerator: There cannot be a swap move given a permutation of length 1.", "permutation");

      //int[] affineFacilities = new int[permutation.Length];
      double totalSum = 0;
      for (int i = 0; i < length; i++) {
        //affineFacilities[i] = ProportionalSelect(random, weightsRowSum, weights, i);
        totalSum += weightsRowSum[i];
      }

      for (int i = 0; i < length - 1; i++)
        for (int j = i + 1; j < length; j++) {
          AffinitySwapMove move = new AffinitySwapMove();
          move.AddSwap(i, j);
          if (weightsRowSum[i] / totalSum >= random.NextDouble()
            || weightsRowSum[j] / totalSum >= random.NextDouble()) {
            int sI = ProportionalSelect(random, weightsRowSum, weights, i), sJ = ProportionalSelect(random, weightsRowSum, weights, j);
            if (!(sI == sJ || sJ == i || sJ == j || sI == i || sI == j))
              move.AddSwap(sI, sJ);
          }
          yield return move;
        }
    }

    protected static int ProportionalSelect(IRandom random, DoubleArray sums, DoubleMatrix matrix, int row) {
      double total = sums[row];
      double rand = random.NextDouble() * total;
      for (int i = 0; i < matrix.Columns; i++) {
        total -= matrix[row, i];
        if (total <= rand) return i;
      }
      throw new InvalidOperationException("StochasticAffinitySwapMoveGenerator: Could not generate moves for any affine item.");
    }
  }
}
